﻿using System;
using System.Collections.Generic;	//List
using System.IO;	//MemoryStream
using System.Threading;		//Sleep
using ModifiedSerialPort;

namespace lcdST7735com
{
	//ST7735のコマンド
	struct LCDCmd
	{
		//System function Command List and Description
		public const byte
			NOP		= 0x00,	//No Operation
			SWRESET	= 0x01,	//Software reset
			SLPIN	= 0x10,	//Sleep in & booster off
			SLPOUT	= 0x11,	//Sleep out & booster on
			PTLON	= 0x12,	//Partial mode on
			NORON	= 0x13,	//Partial off (Normal mode on)
			INVOFF	= 0x20,	//Display inversion off
			INVON	= 0x21,	//Display inversion on
			GAMSET	= 0x26,	//Gamma curve select
			DISPOFF	= 0x28,	//Display off
			DISPON	= 0x29,	//Display on
			CASET	= 0x2A,	//Column address set
			RASET	= 0x2B,	//Row address set
			RAMWR	= 0x2C,	//Memory write
			PTLAR	= 0x30,	//Partial start/end address set
			TEOFF	= 0x34,	//Tearing effect line off
			TEON	= 0x35,	//Tearing effect mode set & on
			MADCTL	= 0x36,	//Memory data access control
			IDMOFF	= 0x38,	//Idle mode off
			IDMON	= 0x39,	//Idle mode on
			COLMOD	= 0x3A;	//Interface pixel format

		//Panel Function Command List and Description
		public const byte
			FRMCTR1	= 0xB1,	//In normal mode (Full colors)
			FRMCTR2	= 0xB2,	//In Idle mode (8-colors)
			FRMCTR3	= 0xB3,	//In partial mode + Full colors
			INVCTR	= 0xB4,	//Display inversion control
			DISSET5	= 0xB6,	//Display function setting
			PWCTR1	= 0xC0,	//Power control setting
			PWCTR2	= 0xC1,	//Power control setting
			PWCTR3	= 0xC2,	//Power control setting / In normal mode (Full colors)
			PWCTR4	= 0xC3,	//Power control setting / In Idle mode (8-colors)
			PWCTR5	= 0xC4,	//Power control setting / In partial mode + Full colors
			VMCTR1	= 0xC5,	//VCOM control 1
			VMOFCTR	= 0xC7,	//Set VCOM offset control
			WRID2	= 0xD1,	//Set LCM version code
			WRID3	= 0xD2,	//Customer Project code
			NVCTR1	= 0xD9,	//EEPROM control status
			NVCTR2	= 0xDE,	//EEPROM Read Command
			NVCTR3	= 0xDF,	//EEPROM Write Command
			GAMCTRP1	= 0xE0,	//Set Gamma adjustment ('+' polarity)
			GAMCTRN1	= 0xE1,	//Set Gamma adjustment ('-' polarity)
			EXTCTRL	= 0xF0,	//Extension Command Control
			PWCTR6	= 0xFC,	//Power control setting / In partial mode + Idle
			VCOM4L	= 0xFF;	//Vcom 4 Level control
	};

	//通信コマンド（文字）
	enum COMCmd
	{							//引数
		LCD_CMD			= '#',	//LCDコマンド[1バイト]
		LCD_BYTE		= '$',	//LCDデータ[1バイト]
		LCD_WORD_RLE	= '*',	//LCDデータ[1ワードHi][Lo], 繰り返し回数[1ワードHi][Lo]
		WHO				= '@',	//なし（デバイス名の確認）
		//
		CMD_BINMODE		= '[',	//転送バイト数[1バイト]（バイナリ転送モードに切り替える）
		LCD_BINDATA		= ']',	//バイナリデータ[nバイト]	※DATA_BUF_SIZE以下の長さのデータ

		RET_UNKNOWN_CMD = '?',	//解釈できないコマンドを受信した場合に送り返す文字（デバイスが）
		RET_RCVERROR = RET_UNKNOWN_CMD	//受信エラー発生（データ不着）の合図として返す（デバイスが）
	};

	class ST7735Ctrl
	{
		//インターフェイスのデバイスに埋め込まれている名前
		private const string DEVICE_NAME = "ST7735";

		//インターフェイスのデバイスのバッファに格納できるデータ長(byte)
		private const int DATA_BUF_SIZE = 96;
		
		//1回の通信の終了コード
		private const char END_CODE = '\n';
		private static readonly char[] TXD_END_CODE = { END_CODE };

		private MSerialPort comCtrl = null;
		public static bool IsComError { get; private set; }	//通信エラーを通知するフラグ

		//LCDの表示サイズ
		private const int
			SCREEN_OFFSET_X = 2,
			SCREEN_OFFSET_Y = 3;
		public const int
			SCREEN_WIDTH = 128,
			SCREEN_HEIGHT = 128;

		/*	座標系と表示オフセット
		コントローラの座標系(128x160または132x162)
			[0,  0]--[131,  0]
			  :			 :
			[0,161]--[131,161]

		ZY-FGD1442701V1-LCDの座標系(128x128)	※コントローラの[2,3]がLCD座標系の[0,0]
			[0,  0]--[127,  0]
			  :          :
			[0,127]--[127,127]
		*/

		//コンストラクタ
		public ST7735Ctrl()
		{
			IsComError = false;
		}
		
		//COMポートオブジェクトをセットする
		public void SetSerialPort(MSerialPort msp)
		{
			comCtrl = msp;
		}

		//指定されたCOMポートにデバイスが接続されているか確認する
		public bool DeviceContact(MSerialPort testComPort)
		{
			if (testComPort == null || !testComPort.IsOpen) return false;
			MSerialPort orgComCtrl = comCtrl;
			comCtrl = testComPort;

			Command(COMCmd.WHO);
			string s = ReadString();
			bool isOK = (!IsComError && s == DEVICE_NAME);

			comCtrl = orgComCtrl;
			return isOK;
		}
		
		//引数の下位4bitの数値0～15にゲタ'A'を履かせて文字に変換する
		private byte itoc(int n) { return (byte)('A' + (n & 0x0F)); }

		//ゲタ'A'を履かせた文字から元の数値0～15に変換する
		private byte ctoi(byte c) { return (byte)(c - 'A'); }

		//コマンドとパラメータを送信する
		//・第2引数は可変長配列。または省略可。
		//・データは、1byteの値を上位／下位4bitに分け、それを加工した文字として送る。
		private void Command(COMCmd cmd, params byte[] values)
		{
			IsComError = false;
			byte[] byteArray = null;
			
			using (MemoryStream memStream = new MemoryStream())
			{
				memStream.WriteByte((byte)cmd);
				
				for (int i = 0; i < values.Length; i++)
				{
					memStream.WriteByte(itoc(values[i] >> 4));
					memStream.WriteByte(itoc(values[i]));
				}
				byteArray = memStream.ToArray();
			}

			try
			{
				comCtrl.ReadExisting();	//受信に備え、PC側の受信バッファをクリアしておく
				comCtrl.Write(byteArray, 0, byteArray.Length);
				comCtrl.Write(TXD_END_CODE, 0, TXD_END_CODE.Length);
			}
			catch { IsComError = true; }
		}

		//送信結果を文字列として受信する
		private string ReadString()
		{
			IsComError = false;
			string s = null;
			try { s = comCtrl.ReadLine().Trim(); }	//Trim()は保険
			catch { IsComError = true; }	//タイムアウト発生はキャッチされる
			return s;
		}
		
		//書き込み系コマンドに対するデバイスからの返信の妥当性をチェックする
		//・書き込み系コマンドを送信すると、コマンド自身を表す文字が返ってくる。
		private bool WriteCommandIsDone(COMCmd cmd)
		{
			string sRetData = ReadString();
			if (IsComError) return false;
			string sWrCmd = ((char)cmd).ToString();
			if (sWrCmd != sRetData) Console.WriteLine(sRetData);	//debug
			return (sWrCmd == sRetData);
		}

		//LCDコマンドを送信する
		//・第2引数は可変長配列。または省略可。
		private bool SendLcdCommand(byte cmd, params byte[] values)
		{
			byte[] args = { cmd };
			Command(COMCmd.LCD_CMD, args);
			bool isOK = WriteCommandIsDone(COMCmd.LCD_CMD);
			
			for (int i = 0; i < values.Length && isOK; i++)
				isOK = isOK && SendLcdData(values[i]);
			
			return isOK;
		}

		//LCDデータを送信する（1バイトのみ）
		private bool SendLcdData(byte data)
		{
			byte[] args = { data };
			Command(COMCmd.LCD_BYTE, args);
			bool isOK = WriteCommandIsDone(COMCmd.LCD_BYTE);
			return isOK;
		}
		
		//LCDデータを送信する（ワードデータをRLE形式で）
		private bool SendLcdData(int data, int length)
		{
			if (length <= 0 || 0xFFFF < length) return false;

			byte[] args = { itobH(data), itobL(data), itobH(length), itobL(length) };
			Command(COMCmd.LCD_WORD_RLE, args);
			bool isOK = WriteCommandIsDone(COMCmd.LCD_WORD_RLE);
			return isOK;
		}

		private byte itobH(int n) { return (byte)((n >> 8) & 0xFF); }
		private byte itobL(int n) { return (byte)(n & 0xFF); }
		
		//LCDデータを送信する（バイナリ形式で）
		private bool SendLcdBinData(byte[] byteArray)
		{
			if (byteArray == null || byteArray.Length == 0 || DATA_BUF_SIZE < byteArray.Length) return false;
			byte[] args = { (byte)byteArray.Length };
			Command(COMCmd.CMD_BINMODE, args);
			bool isOK = WriteCommandIsDone(COMCmd.CMD_BINMODE);
			if (!isOK) return false;

			try
			{
				comCtrl.ReadExisting();	//受信に備え、PC側の受信バッファをクリアしておく
				comCtrl.Write(byteArray, 0, byteArray.Length);
			}
			catch { IsComError = true; }

			isOK = WriteCommandIsDone(COMCmd.LCD_BINDATA);
			return isOK;
		}

		//ZY-FGD1442701V1-LCDの初期化
		public void Init()
		{
			SendLcdCommand(LCDCmd.SWRESET);	//Software reset:	Sleep-In, Display-Off
			Thread.Sleep(120);
			SendLcdCommand(LCDCmd.SLPOUT);	//Sleep out & booster on
			Thread.Sleep(120);
			
#if false
			//aitendoサンプルコードより
			SendLcdCommand(0xFF, 0x40, 0x03, 0x1A);	//Vcom 4 Level control
			SendLcdCommand(0xd9, 0x60, 0xc7, 0x90);	Thread.Sleep(200);	//EEPROM control status
			SendLcdCommand(0xB1, 0x04, 0x25, 0x18);	//In normal mode (Full colors)
			SendLcdCommand(0xB2, 0x04, 0x25, 0x18);	//In Idle mode (8-colors)
			SendLcdCommand(0xB3, 0x04, 0x25, 0x18, 0x04, 0x25, 0x18);	//In partial mode + Full colors
			SendLcdCommand(0xB4, 0x03);	//Display inversion control
			SendLcdCommand(0xB6, 0x15, 0x02);	//Display function setting
			SendLcdCommand(0xC0, 0x02, 0x70);	//Power control setting
			SendLcdCommand(0xC1, 0x07);	//Power control setting
			SendLcdCommand(0xC2, 0x01, 0x01);	//Power control setting / In normal mode (Full colors)
			SendLcdCommand(0xC3, 0x02, 0x07);	//Power control setting / In Idle mode (8-colors)
			SendLcdCommand(0xC4, 0x02, 0x04);	//Power control setting / In partial mode + Full colors
			SendLcdCommand(0xFC, 0x11, 0x17);	//Power control setting / In partial mode + Idle
			SendLcdCommand(0xC5, 0x3c, 0x4f);	//VCOM control 1
			SendLcdCommand(0x36, 0xC8);	//Memory data access control
			SendLcdCommand(0x3a, 0x05);	//Interface pixel format

			SendLcdCommand(0xE0,
				0x06,0x0E,0x05,0x20,0x27,0x23,0x1C,0x21,0x20,0x1C,0x26,0x2F,0x00,0x03,0x00,0x24
				);	//Set Gamma adjustment ('+' polarity)
			SendLcdCommand(0xE1,
				0x06,0x10,0x05,0x21,0x27,0x22,0x1C,0x21,0x1F,0x1D,0x27,0x2F,0x05,0x03,0x00,0x3F
				);	//Set Gamma adjustment ('-' polarity)

			SendLcdCommand(0x2A, 0x00, 0x02, 0x00, 0x81);	//Column address set
			SendLcdCommand(0x2B, 0x00, 0x03, 0x00, 0x82);	//Row address set

			SendLcdCommand(0x29);	//Display on
			Thread.Sleep(120);
#else
			//コントローラと液晶画面の接続情報
			SendLcdCommand(LCDCmd.MADCTL, 0xC8);	//Memory data access control
			/*
			bit7(MY)	0:順方向走査, 1:逆方向走査（垂直反転）
			bit6(MX)	0:順方向走査, 1:逆方向走査（水平反転）
			bit5(MV)	0:水平方向優先走査, 1:垂直方向優先走査（90度回転）
			bit4(ML)	Vertical Refresh Order
			bit3(RGB)	0:RGB, 1:BGR（メモリ上のRGBデータに対する液晶画面側の解釈）
			bit2(MH)	Horizontal Refresh Order
			bit1	なし
			bit0	なし

			【ML,MHについて】
			画面更新時、全体がパッと変わるわけではない。ワイプ効果のように画面端からスーッと変わっていく。
			MLは垂直方向どちら向き書き換わっていくかを決める。0:上→下, 1:下→上
			MHは水平方向どちら向き書き換わっていくかを決める。0:左→右, 1:右→左

			【ZY-FGD1442701V1-LCDの場合】
			LCDモジュールの端子を下側にして見た場合、「MY,MX」とも反転すると画面左上が座標原点となる。
			bit3を「RGB」とした場合、LCDモジュール上はBGR順で解釈した色で表示される。
			よって、bit3は「BGR」とする。
			以上より、パラメータは「0xC8」となる。
			*/

			//色深度
			SendLcdCommand(LCDCmd.COLMOD, 0x05);	//Interface pixel format
			//0x03:	12-bit/pixel	RGB=4:4:4bit
			//0x05:	16-bit/pixel	RGB=5:6:5bit
			//0x06:	18-bit/pixel	RGB=6:6:6bit

			Clear();

			SendLcdCommand(LCDCmd.DISPON);	//Display on
#endif
		}

		//描画領域を指定する
		//戻り値:	実際にセットされた描画領域のドット数。不正な引数のときは0
		//・(x,y)を起点とし、この点を含み、幅w高さhの長方形領域を指定する。
		//・有効画面領域を超える分はトリミングする。
		private int SetRect(int x, int y, int w, int h)
		{
			if (x < 0 || SCREEN_WIDTH <= x || y < 0 || SCREEN_HEIGHT <= y || w <= 0 || h <= 0) return 0;
			
			if (SCREEN_WIDTH < x + w) w = SCREEN_WIDTH - x;
			if (SCREEN_HEIGHT < y + h) h = SCREEN_HEIGHT - y;
			x += SCREEN_OFFSET_X;
			y += SCREEN_OFFSET_Y;
			
			int xStart = x, xEnd = x + w - 1;	//開始座標, 終了座標
			int yStart = y, yEnd = y + h - 1;	//開始座標, 終了座標
			SendLcdCommand(LCDCmd.CASET, itobH(xStart), itobL(xStart), itobH(xEnd), itobL(xEnd));
			SendLcdCommand(LCDCmd.RASET, itobH(yStart), itobL(yStart), itobH(yEnd), itobL(yEnd));
			
			return w * h;
		}

		//指定色で領域を塗りつぶす
		//引数	color: 16bitの色値(RGB=5:6:5bit)
		public void FillRect(int x, int y, int w, int h, int color)
		{
			int size = SetRect(x, y, w, h);
			if (size == 0) return;
			SendLcdCommand(LCDCmd.RAMWR);
			SendLcdData(color, size);
		}

		//描画を消去する
		//・任意の色で画面全域を塗りつぶす。
		public void Clear() { Clear(0x0000); }	//0x0000=黒
		public void Clear(int color)
		{
			FillRect(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, color);
		}

		//指定色で点を描画する
		//引数	color: 16bitの色値(RGB=5:6:5bit)
		public void SetPixel(int x, int y, int color)
		{
			FillRect(x, y, 1, 1, color);
		}

		//画像を描画する
		public void	DrawImage(string filename)
		{
			if (string.IsNullOrEmpty(filename)) return;
			
			ImageCtrl imgCtrl = new ImageCtrl(filename);
			if (imgCtrl.IsError) return;
			if (imgCtrl.ImgW < imgCtrl.ImgH)
				imgCtrl.Resize(0, SCREEN_HEIGHT);
			else
				imgCtrl.Resize(SCREEN_WIDTH, 0);
			
			int[] pixels = imgCtrl.GetPixelArray();
			
			int posX = (SCREEN_WIDTH - imgCtrl.ImgW) / 2;
			int posY = (SCREEN_HEIGHT - imgCtrl.ImgH) / 2;
			DrawImage(posX, posY, imgCtrl.ImgW, imgCtrl.ImgH, pixels);
		}

		//画像を描画する
		//引数	x,y,w,h:	描画領域（画像の縦横サイズと一致していること）
		//		pixels:		配列の要素は16bit色値
		private void DrawImage(int x, int y, int w, int h, int[] pixels)
		{
			if (pixels == null || pixels.Length == 0) return;

			int size = SetRect(x, y, w, h);
			if (size == 0) return;

			SendLcdCommand(LCDCmd.RAMWR);

			//1画素あたり2byteの色データで表されている
			//・マイコン側の受信バッファのサイズは DATA_BUF_SIZEバイトなので、
			//	一度に送信できる画素数は (DATA_BUF_SIZE / 2)ピクセル。
			byte[] byteAry = new byte[DATA_BUF_SIZE];
			int MaxCopyPixelLength = DATA_BUF_SIZE / 2;
			bool isOK = true;
			for (int pos = 0; pos < pixels.Length && isOK; pos += MaxCopyPixelLength)
			{
				int copyPixelLength = Math.Min(pixels.Length - pos, MaxCopyPixelLength);
				if (copyPixelLength < MaxCopyPixelLength) byteAry = new byte[copyPixelLength * 2];
				for (int a = 0, i = 0; i < copyPixelLength; i++)
				{
					byteAry[a++] = itobH(pixels[pos + i]);
					byteAry[a++] = itobL(pixels[pos + i]);
				}

				int retry = 0;
				do { isOK = SendLcdBinData(byteAry); }
				while (!isOK && ++retry < 10);
			}
			if (!isOK) Console.WriteLine("error break.");
		}
		
	}	//class
}	//namespace
